package net.sf.logdistiller;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.util.*;

import net.sf.logdistiller.publishers.BasePublishers;
import net.sf.logdistiller.util.ExtensionHelper;
import net.sf.logdistiller.util.FormatUtil;

/**
 * Definition of available report publishers.
 */
public abstract class Publishers
{
    /**
     * get the <code>Publishers</code> instances defined by this extension.
     *
     * @return the list of publishers
     * @see LogType
     */
    public abstract List definePublishers();

    public static final List ALL;

    public static final Map MAP;
    static
    {
        List all = new ArrayList( new BasePublishers().definePublishers() );
        all.addAll( loadExtensionPublishers() );
        ALL = Collections.unmodifiableList( all );

        Map map = new HashMap();
        Iterator iter = ALL.iterator();
        while ( iter.hasNext() )
        {
            Publisher type = (Publisher) iter.next();
            map.put( type.getId(), type );
        }
        MAP = Collections.unmodifiableMap( map );
    }

    /**
     * get all the report publishers defined (predefined and custom).
     */
    public static List getAllPublishers()
    {
        return ALL;
    }

    public static String listAllPublishersIds()
    {
        return FormatUtil.join( ", ", MAP.keySet().iterator() );
    }

    public static Publisher getPublisher( String id )
    {
        return (Publisher) MAP.get( id );
    }

    /**
     * Loads report publishers defined by extension mechanism: if defined, <code>publishers</code> property in
     * <code>logdistiller.properties</code> is the full class name of a concrete implementation of
     * <code>Publishers</code> class.
     *
     * @return List the list of every <code>Publisher</code> defined by loaded <code>Publishers</code>
     */
    private static List loadExtensionPublishers()
    {
        Iterator iter = ExtensionHelper.findExtensions( "publishers" ).iterator();
        List publishers = new ArrayList();
        while ( iter.hasNext() )
        {
            String publishersClass = (String) iter.next();
            if ( publishersClass != null )
            {
                publishers.addAll( loadPublishers( publishersClass ) );
            }
        }
        return publishers;
    }

    private static List loadPublishers( String publishersClass )
    {
        try
        {
            Publishers publishers = (Publishers) Class.forName( publishersClass ).newInstance();
            return publishers.definePublishers();
        }
        catch ( ClassNotFoundException cnfe )
        {
            throw new RuntimeException( "unable to load plugins class " + publishersClass, cnfe );
        }
        catch ( IllegalAccessException iae )
        {
            throw new RuntimeException( "unable to access plugins constructor for class " + publishersClass, iae );
        }
        catch ( InstantiationException ie )
        {
            throw new RuntimeException( "unable to instanciate plugins class " + publishersClass, ie );
        }
    }
}
