package net.sf.logdistiller;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.util.*;

import net.sf.logdistiller.reports.BaseReportFormats;
import net.sf.logdistiller.util.ExtensionHelper;
import net.sf.logdistiller.util.FormatUtil;

/**
 * Definition of available report formats. By default, some report formats come
 * {@link net.sf.logdistiller.reports.BaseReportFormats predefined} with <b>LogDistiller</b>, but you can add your own.
 */
public abstract class ReportFormats
{
    public abstract List defineReportFormats();

    public static final List ALL;

    public static final Map MAP;
    static
    {
        List all = new ArrayList( new BaseReportFormats().defineReportFormats() );
        all.addAll( loadExtensionReportFormats() );
        ALL = Collections.unmodifiableList( all );

        Map map = new HashMap();
        Iterator iter = ALL.iterator();
        while ( iter.hasNext() )
        {
            ReportFormat reportFormat = (ReportFormat) iter.next();
            map.put( reportFormat.getId(), reportFormat );
        }
        MAP = Collections.unmodifiableMap( map );
    }

    /**
     * get all the plugins defined (predefined and custom).
     */
    public static List getAllReportFormats()
    {
        return ALL;
    }

    public static String listAllReportFormatIds()
    {
        return FormatUtil.join( ", ", MAP.keySet().iterator() );
    }

    public static ReportFormat getReportFormat( String id )
    {
        ReportFormat format = (ReportFormat) MAP.get( id );
        if ( format == null )
        {
            throw new IllegalArgumentException( "unknown report format '" + id + "', known values are: "
                + listAllReportFormatIds() );
        }
        return format;
    }

    private static List loadExtensionReportFormats()
    {
        Iterator iter = ExtensionHelper.findExtensions( "reportformats" ).iterator();
        List reportformats = new ArrayList();
        while ( iter.hasNext() )
        {
            String reportformatsClass = (String) iter.next();
            if ( reportformatsClass != null )
            {
                reportformats.addAll( loadReportFormats( reportformatsClass ) );
            }
        }
        return reportformats;
    }

    private static List loadReportFormats( String reportFormatsClass )
    {
        try
        {
            ReportFormats reportFormats = (ReportFormats) Class.forName( reportFormatsClass ).newInstance();
            return reportFormats.defineReportFormats();
        }
        catch ( ClassNotFoundException cnfe )
        {
            throw new RuntimeException( "unable to load report formats class " + reportFormatsClass, cnfe );
        }
        catch ( IllegalAccessException iae )
        {
            throw new RuntimeException( "unable to access report formats constructor for class " + reportFormatsClass,
                                        iae );
        }
        catch ( InstantiationException ie )
        {
            throw new RuntimeException( "unable to instanciate report formats class " + reportFormatsClass, ie );
        }
    }
}
