package net.sf.logdistiller.gui;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.awt.Insets;
import java.io.InputStream;
import java.util.*;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.jgoodies.forms.layout.CellConstraints;

import net.sf.logdistiller.xml.DOMUtils;

/**
 * This class is a simplified version of FormLayoutMaker's equivalent class, in a dream to avoid need for Java 1.4 at
 * runtime.
 */
public class SimpleLayoutConstraintsManager
{
    private final Map layouts = new HashMap();

    private SimpleLayoutConstraintsManager()
    {
    }

    public SimpleContainerLayout getLayout( String name )
    {
        return (SimpleContainerLayout) layouts.get( name );
    }

    public void addLayout( SimpleContainerLayout containerLayout )
    {
        layouts.put( containerLayout.getName(), containerLayout );
    }

    public static final String DEFAULT = "default";

    private final static Map CELL_CONSTRAINTS = new HashMap()
    {
        {
            put( "default", CellConstraints.DEFAULT );
            put( "fill", CellConstraints.FILL );
            put( "center", CellConstraints.CENTER );
            put( "left", CellConstraints.LEFT );
            put( "right", CellConstraints.RIGHT );
            put( "top", CellConstraints.TOP );
            put( "bottom", CellConstraints.BOTTOM );
        }
    };

    public static CellConstraints.Alignment getAlignment( String value )
    {
        CellConstraints.Alignment alignment =
            (CellConstraints.Alignment) CELL_CONSTRAINTS.get( ( value == null ) ? DEFAULT : value );
        if ( alignment == null )
        {
            throw new RuntimeException( "Invalid alignment '" + value + "'" );
        }
        return alignment;
    }

    public static SimpleLayoutConstraintsManager getLayoutConstraintsManager( InputStream stream )
    {
        try
        {
            DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            Document dataDocument = documentBuilder.parse( stream );
            Element root = (Element) dataDocument.getDocumentElement();
            return getLayoutConstraintsManager( root );
        }
        catch ( Exception e )
        {
            throw new RuntimeException( "Unable to create DocumentBuilder", e );
        }
    }

    public static SimpleLayoutConstraintsManager getLayoutConstraintsManager( Element containersNode )
    {
        if ( !containersNode.getNodeName().equals( "containers" ) )
        {
            throw new RuntimeException( "Expected a node named 'containers', found '" + containersNode.getNodeName()
                + "'" );
        }
        SimpleLayoutConstraintsManager layoutConstraintsManager = new SimpleLayoutConstraintsManager();
        Iterator iter = DOMUtils.getChildElementsByTagName( containersNode, "container" );
        while ( iter.hasNext() )
        {
            Element containerNode = (Element) iter.next();
            String containerName = containerNode.getAttribute( "name" );
            String columnSpecs = containerNode.getAttribute( "columnSpecs" );
            String rowSpecs = containerNode.getAttribute( "rowSpecs" );
            SimpleContainerLayout containerLayout = new SimpleContainerLayout( containerName, columnSpecs, rowSpecs );

            Iterator iterConstraints = DOMUtils.getChildElementsByTagName( containerNode, "cellconstraints" );
            while ( iterConstraints.hasNext() )
            {
                Element constraintsNode = (Element) iterConstraints.next();
                String name = constraintsNode.getAttribute( "name" );
                CellConstraints.Alignment horizontalAlignment =
                    getAlignment( constraintsNode.getAttribute( "horizontalAlignment" ) );
                CellConstraints.Alignment verticalAlignment =
                    getAlignment( constraintsNode.getAttribute( "verticalAlignment" ) );
                int gridX = getIntValue( constraintsNode, "gridX", 1 );
                int gridY = getIntValue( constraintsNode, "gridY", 1 );
                int gridWidth = getIntValue( constraintsNode, "gridWidth", 1 );
                int gridHeight = getIntValue( constraintsNode, "gridHeight", 1 );
                int topInset = getIntValue( constraintsNode, "topInset", 0 );
                int bottomInset = getIntValue( constraintsNode, "bottomInset", 0 );
                int rightInset = getIntValue( constraintsNode, "rightInset", 0 );
                int leftInset = getIntValue( constraintsNode, "leftInset", 0 );

                CellConstraints constraints =
                    new CellConstraints( gridX, gridY, gridWidth, gridHeight, horizontalAlignment, verticalAlignment,
                                         new Insets( topInset, leftInset, bottomInset, rightInset ) );
                containerLayout.addCellConstraints( name, constraints );
            }
            layoutConstraintsManager.addLayout( containerLayout );
        }
        return layoutConstraintsManager;
    }

    private static int getIntValue( Element elmt, String name, int defaultValue )
    {
        String value = elmt.getAttribute( name );
        return ( value == null ) ? defaultValue : Integer.parseInt( value );
    }
}
