package net.sf.logdistiller.logtypes;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.*;
import java.text.ParseException;

import net.sf.logdistiller.LogEvent;
import net.sf.logdistiller.LogType;
import net.sf.logdistiller.util.StringCutter;

/**
 * Log event for Unix syslog facility. By default, the classification rules generated by the GUI for this type of logs
 * will sort events based on the following attributes: <code>host</code>, then <code>program</code>.
 */
public class SyslogLogEvent
    extends LogEvent
{
    public final static String ID = "syslog";

    // see http://logreport.org/doc/gen/os/
    public final String timestamp;

    public final String host;

    public final String program;

    public final String pid;

    public final String message;

    public final static LogType LOGTYPE = new LogType.Basic( ID );

    private final static String[] ATTRIBUTE_NAMES = { "logSource", "timestamp", "host", "program", "pid", "message" };

    public final static LogType.Description DESCRIPTION = new Description( (LogType.Basic) LOGTYPE, ATTRIBUTE_NAMES );

    public SyslogLogEvent( Factory factory, String rawLog )
        throws ParseException
    {
        super( factory, rawLog );
        timestamp = rawLog.substring( 0, 15 );
        StringCutter sc = new StringCutter( rawLog, 16 );
        host = sc.parseTo( " " );
        String remaining = sc.getRemaining();
        if ( remaining.indexOf( ": " ) < 0 )
        {
            program = "";
            pid = "";
            message = remaining;
        }
        else
        {
            // <program>[<pid>]: <message>
            String programPid = sc.parseTo( ": " );
            message = sc.getRemaining();

            int index = programPid.indexOf( '[' );
            if ( ( index >= 0 ) && programPid.endsWith( "]" ) )
            {
                program = programPid.substring( 0, index );
                pid = programPid.substring( index + 1, programPid.length() - 1 );
            }
            else
            {
                program = programPid;
                pid = "";
            }
        }
        setAttributes( new String[] { factory.getLogSource(), timestamp, host, program, pid, message } );
    }

    private static class Description
        extends LogType.Description
    {
        public Description( LogType.Basic logtype, String[] attributeNames )
        {
            super( logtype, attributeNames );
            logtype.setDescription( this );
        }

        public LogEvent.Factory newFactory( Reader reader, String logSource )
            throws IOException
        {
            return new Factory( this, reader, logSource );
        }

        public String getDefaultSamplingAttributes()
        {
            return "host,program";
        }
    }

    private static class Factory
        extends LogEvent.Factory
    {
        private final LineNumberReader reader;

        public Factory( Description description, Reader reader, String logSource )
            throws FileNotFoundException
        {
            super( description, logSource );
            this.reader = new LineNumberReader( reader );
        }

        protected LogEvent readNextEvent()
            throws IOException, ParseException
        {
            String curLine = reader.readLine();
            if ( curLine == null )
            {
                // EOF
                return null;
            }
            try
            {
                return new SyslogLogEvent( this, curLine );
            }
            catch ( RuntimeException re )
            {
                throw new RuntimeException( "error while parsing line '" + curLine + "'", re );
            }
        }
    }
}
