package net.sf.logdistiller.plugins;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.*;

import net.sf.logdistiller.LogDistillation;
import net.sf.logdistiller.LogDistiller;
import net.sf.logdistiller.LogEvent;
import net.sf.logdistiller.Plugin;
import net.sf.logdistiller.ReportFormat;
import net.sf.logdistiller.util.FormatUtil;

/**
 * <b>Log sampling</b> plugin: works like FreqPlugin, but saves some log events to a log file.
 * <p>
 * Parameters (added to FreqPlugin's parameters):
 * </p>
 * <ul>
 * <li><code>sampling.filename</code> (default: <code>[id]-sampling.log</code>): name of the file to save the sample of
 * log events for each attribute's value.</li>
 * <li><code>sampling.maxCount</code> (default: 10, -1 = no limit): number of log events to save for each attribute's
 * value.</li>
 * <li><code>sampling.maxSize</code> (default: -1 = no limit): maximum size (in kb) of the sample file</li>
 * </ul>
 *
 * @see FreqPlugin
 * @since 0.7
 */
public class SamplingPlugin
    extends FreqPlugin
{
    private static final long serialVersionUID = -2960554427849032861L;

    public final static String ID = "sampling";

    public final static Plugin TYPE = new Type();

    private final int samplingMaxCount;

    private final long samplingMaxSize;

    private final String samplingFilename;

    private File samplingLogFile;

    private transient PrintStream samplingLogStream;

    public SamplingPlugin( LogDistiller.Plugin definition )
    {
        super( definition, false );

        samplingMaxCount = Integer.parseInt( definition.getParam( "sampling.maxCount", "10" ) );
        samplingMaxSize = Long.parseLong( definition.getParam( "sampling.maxSize", "-1" ) ) * 1024;
        samplingFilename = definition.getParam( "sampling.filename", definition.getGroup().getId() + "-sampling.log" );
    }

    public void begin( File destinationDirectory )
        throws FileNotFoundException
    {
        super.begin( destinationDirectory );
        samplingLogFile = new File( destinationDirectory, samplingFilename );
        samplingLogStream = new PrintStream( new FileOutputStream( samplingLogFile ) );
    }

    public void addLogEventToFreq( LogEvent logEvent, Freq freq )
        throws IOException
    {
        super.addLogEventToFreq( logEvent, freq );

        if ( ( ( samplingMaxCount < 0 ) || ( ( samplingMaxCount > 0 ) && ( freq.count <= samplingMaxCount ) ) )
            && ( ( samplingMaxSize < 0 ) || ( ( samplingMaxSize > 0 ) && ( freq.bytes <= samplingMaxSize ) ) ) )
        {
            samplingLogStream.println( logEvent.getRawLog() );
        }
    }

    public void end()
        throws IOException
    {
        samplingLogStream.close();
        super.end();
    }

    protected void appendLinkToFreqReport( ReportFormat.PluginReport report )
    {
        report.addLink( samplingLogFile.getName(), "sampling log file (max count = "
            + ( ( samplingMaxCount < 0 ) ? "none" : String.valueOf( samplingMaxCount ) ) + ", max size = "
            + ( ( samplingMaxSize < 0 ) ? "none" : FormatUtil.formatSize( samplingMaxSize ) ) + ")" );
    }

    private static class Type
        extends Plugin
    {
        public Type()
        {
            super( ID );
        }

        public LogDistillation.Plugin newInstance( LogDistiller.Plugin conf )
        {
            return new SamplingPlugin( conf );
        }
    }
}
