package net.sf.logdistiller.util;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Set;

public class FormatUtil
{
    private FormatUtil()
    {
    }

    public static String formatPeriod( final long periodMs )
    {
        long base = ( periodMs < 60000 ) ? periodMs : ( periodMs + 500 );
        int milliseconds = (int) ( base % 1000 );
        long remainder = base / 1000;
        int seconds = (int) ( remainder % 60 );
        remainder /= 60;
        int minutes = (int) ( remainder % 60 );
        remainder /= 60;
        int hours = (int) ( remainder % 24 );
        remainder /= 24;
        int days = (int) remainder;
        StringBuffer buffer = new StringBuffer();
        int fieldCount = 0;
        if ( days > 0 )
        {
            fieldCount++;
            buffer.append( days + " d" );
        }
        if ( ( fieldCount > 0 ) || ( hours > 0 ) )
        {
            buffer.append( ( ( fieldCount > 0 ) ? " " : "" ) + hours + " h" );
            fieldCount++;
        }
        if ( ( fieldCount > 0 ) || ( minutes > 0 ) )
        {
            buffer.append( ( ( fieldCount > 0 ) ? " " : "" ) + minutes + " m" );
            fieldCount++;
        }
        if ( ( fieldCount > 0 ) || ( seconds > 0 ) )
        {
            buffer.append( ( ( fieldCount > 0 ) ? " " : "" ) + seconds + " s" );
            fieldCount++;
        }
        if ( fieldCount < 2 )
        {
            buffer.append( ( ( fieldCount > 0 ) ? " " : "" ) + milliseconds + " ms" );
        }
        return buffer.toString();
    }

    public static String formatSize( final long size )
    {
        if ( size < 1024 )
        {
            return String.valueOf( size ) + " b";
        }
        long sizeKb = size / 1024;
        if ( sizeKb < 1024 )
        {
            return String.valueOf( sizeKb ) + '.' + String.valueOf( ( size % 1024 ) * 10 / 1024 ) + " kb";
        }
        long sizeMb = sizeKb / 1024;
        return String.valueOf( sizeMb ) + '.' + String.valueOf( ( sizeKb % 1024 ) * 10 / 1024 ) + " Mb";
    }

    /**
     * Formats a fraction as a "xx.x%" string.
     *
     * @param part int the portion of the total
     * @param total int the total
     * @return String the value of (fraction / total) as "xx.x%"
     */
    public static String formatFraction( long part, long total )
    {
        int fraction = Math.round( (float) ( part * 1000 ) / total );
        return String.valueOf( fraction / 10 ) + '.' + ( fraction % 10 ) + '%';
    }

    public static String join( String separator, Iterator iter )
    {
        StringBuffer buff = new StringBuffer();
        while ( iter.hasNext() )
        {
            if ( buff.length() > 0 )
            {
                buff.append( separator );
            }
            buff.append( iter.next() );
        }
        return buff.toString();
    }

    public static String join( String separator, Object[] values )
    {
        return join( separator, Arrays.asList( values ).iterator() );
    }

    public static String join( String separator, List list )
    {
        return join( separator, list.iterator() );
    }

    public static String join( String separator, Set set )
    {
        return join( separator, set.iterator() );
    }

    public static Locale parseLocale( String locale )
    {
        int len = locale.length();
        if ( ( len != 2 ) && ( len != 5 ) && ( len != 8 ) )
        {
            throw new RuntimeException( "bad locale format: '" + locale + "', sould be xx?yy?zz" );
        }
        String lang = locale.substring( 0, 2 );
        if ( len == 2 )
        {
            return new Locale( lang );
        }
        String country = locale.substring( 3, 5 );
        if ( len == 5 )
        {
            return new Locale( lang, country );
        }
        String variant = locale.substring( 6 );
        return new Locale( lang, country, variant );
    }
}
