package net.sf.logdistiller.xml;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.*;
import java.util.*;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import net.sf.logdistiller.*;
import net.sf.logdistiller.util.PropertiesReplacer;
import org.xml.sax.ErrorHandler;

/**
 * This class parses an XML rules configuration file and creates corresponding LogDistiller object.
 *
 * @see net.sf.logdistiller.LogDistiller
 */
public class DOMConfigurator
{
    private final Map overriddenProperties;

    private PropertiesReplacer replacer = null;

    private boolean workInProgress = false;

    private boolean old;

    private List warnings = null;

    private LogDistiller ld = null;

    public DOMConfigurator()
    {
        overriddenProperties = null;
    }

    public DOMConfigurator( Map overriddenProperties )
    {
        this.overriddenProperties = overriddenProperties;
    }

    synchronized public LogDistiller read( File f )
        throws ParserConfigurationException, IOException, SAXException
    {
        return read( f, null );
    }

    synchronized public LogDistiller read( File f, ErrorHandler errorHandler )
        throws ParserConfigurationException, IOException, SAXException
    {
        return read( new FileInputStream( f ), errorHandler );
    }

    synchronized public LogDistiller read( InputStream in )
        throws ParserConfigurationException, IOException, SAXException
    {
        return read( in, null );
    }

    synchronized public LogDistiller read( InputStream in, ErrorHandler errorHandler )
        throws ParserConfigurationException, IOException, SAXException
    {
        return read( new InputSource( in ), errorHandler );
    }

    synchronized public LogDistiller read( InputSource inputSource )
        throws ParserConfigurationException, IOException, SAXException
    {
        return read( inputSource, null );
    }

    /**
     * begin configuration, if not currently in progress
     *
     * @return boolean true if really begin, false if configuration is already in progress
     */
    protected boolean begin()
    {
        if ( workInProgress )
        {
            // parse() method is called by read()
            return false;
        }
        workInProgress = true;
        warnings = new ArrayList();
        old = false;
        return true;
    }

    protected void end( boolean really )
    {
        if ( !really )
        {
            return;
        }
        workInProgress = false;
        if ( ld != null )
        {
            Iterator iter = warnings.iterator();
            while ( iter.hasNext() )
            {
                ld.addWarning( (String) iter.next() );
            }
        }
        warnings = null;
        ld = null;
    }

    protected void addWarning( String warning )
    {
        warnings.add( warning );
    }

    synchronized public LogDistiller read( InputSource inputSource, ErrorHandler errorHandler )
        throws ParserConfigurationException, IOException, SAXException
    {
        boolean beginning = begin();
        try
        {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setValidating( true );
            DocumentBuilder docBuilder = dbf.newDocumentBuilder();
            docBuilder.setErrorHandler( ( errorHandler == null ) ? new SAXErrorHandler() : errorHandler );
            LogDistillerEntityResolver resolver = new LogDistillerEntityResolver();
            docBuilder.setEntityResolver( resolver );
            Document doc = docBuilder.parse( inputSource );

            if ( resolver.getCurrentDtd() == null )
            {
                addWarning( "no DTD used: you should use " + LogDistillerEntityResolver.LATEST_DTD_URL );
            }
            else if ( !LogDistillerEntityResolver.LATEST_DTD.equals( resolver.getCurrentDtd() ) )
            {
                addWarning( "using old DTD " + resolver.getCurrentDtd() + ": you should consider upgrading to "
                    + LogDistillerEntityResolver.LATEST_DTD + " (please read upgrade documentation)" );
            }

            ld = parse( doc.getDocumentElement() );
            return ld;
        }
        finally
        {
            end( beginning );
        }
    }

    synchronized public LogDistiller parse( Element element )
    {
        boolean beginning = begin();
        try
        {
            if ( !"logdistiller".equals( element.getNodeName() ) )
            {
                throw new IllegalArgumentException( "document root element is '" + element.getNodeName()
                    + "' where it should be 'logdistiller': see " + LogDistillerEntityResolver.LATEST_DTD_URL );
            }

            String id = element.getAttribute( "id" );
            Map properties = parseProperties( DOMUtils.getChildElementsByTagName( element, "property" ) );
            replacer = new PropertiesReplacer( properties );
            String description = replacer.replaceProperties( DOMUtils.getPCDATAByTagName( element, "description" ) );
            LogDistiller.LogType logtype = null;
            LogDistiller.Output output = null;

            Iterator iter = DOMUtils.getChildElementsByTagName( element, "logtype" );
            if ( iter.hasNext() )
            { // logtype element appeared in logdistiller-1_3.dtd only
                logtype = parseLogType( (Element) iter.next() );
                output = parseOutput( (Element) DOMUtils.getChildElementsByTagName( element, "output" ).next() );
            }
            else
            {
                old = true;
                Map params = parseParams( DOMUtils.getChildElementsByTagName( element, "param" ) );
                String logsUrl = (String) params.get( "logs.url" );
                output =
                    new LogDistiller.Output( "logs", logsUrl, "default", "", params, buildCompatibilityReports( params ) );
            }

            LogDistiller.Category[] categories =
                parseCategories( DOMUtils.getChildElementsByTagName( element, "category" ) );
            LogDistiller.Group[] groups =
                parseGroups( DOMUtils.getChildElementsByTagName( element, "group" ), categories );
            return new LogDistiller( id, description, logtype, output, categories, groups, old );
        }
        finally
        {
            end( beginning );
        }
    }

    protected Map parseProperties( Iterator iter )
    {
        Map map = new HashMap();
        while ( iter.hasNext() )
        {
            Element param = (Element) iter.next();
            String name = param.getAttribute( "name" );
            String value = param.getAttribute( "value" );
            if ( map.put( name, value ) != null )
            {
                addWarning( "multiple values were defined for '" + name
                    + "' property: the last one will override other values." );
            }
        }
        if ( overriddenProperties != null )
        {
            // override properties set by this configurator
            Iterator iter2 = overriddenProperties.keySet().iterator();
            while ( iter2.hasNext() )
            {
                String name = (String) iter2.next();
                if ( map.put( name, overriddenProperties.get( name ) ) == null )
                {
                    throw new IllegalArgumentException( "unable to override property '" + name + "': unknown property" );
                }
            }
        }
        return map;
    }

    protected LogDistiller.LogType parseLogType( Element element )
    {
        String id = element.getAttribute( "id" );
        Map params = parseParams( DOMUtils.getChildElementsByTagName( element, "param" ) );
        String paramAttributes = (String)params.get( "attributes" );

        Attributes attributes = null;
        Iterator iter = DOMUtils.getChildElementsByTagName( element, "attributes" );
        if ( iter.hasNext() )
        { // attributes element appeared in logdistiller-1_4.dtd only
            attributes = parseAttributes( paramAttributes, (Element) iter.next() );
        }
        else
        {
            attributes = new Attributes( paramAttributes, Attributes.NO_EXTENSIONS );
            addWarning( "no logtype/attributes element defined: please consider adding this element (new in LogDistiller 1.1)." );
        }
        return new LogDistiller.LogType( id, params, attributes );
    }

    protected Attributes parseAttributes( String paramAttributes, Element element )
    {
        String provided = DOMUtils.getPCDATAByTagName( element, "provided" );
        if ( provided == null )
        {
            addWarning( "logtype 'attributes' param has been replaced by <provided> element: please consider using this element (new in LogDistiller 1.1)" );
            provided = paramAttributes;
        }
        Attributes.Extension[] extensions = parseExtensions( DOMUtils.getChildElementsByTagName( element, "extension" ) );
        return new Attributes( provided, extensions );
    }

    protected Attributes.Extension[] parseExtensions( Iterator iter )
    {
        List extensions = new ArrayList();
        while ( iter.hasNext() )
        {
            Element extension = (Element) iter.next();
            extensions.add( parseExtension( extension ) );
        }
        return (Attributes.Extension[]) extensions.toArray( new Attributes.Extension[extensions.size()] );
    }

    protected Attributes.Extension parseExtension( Element element )
    {
        String source = element.getAttribute( "source" );
        String provides = element.getAttribute( "provides" );
        String regexp = DOMUtils.getPCDATA( element );
        return new Attributes.Extension( source, provides, regexp );
    }

    protected LogDistiller.Output parseOutput( Element element )
    {
        String directory = replacer.replaceProperties( element.getAttribute( "directory" ) );
        String url = replacer.replaceProperties( element.getAttribute( "url" ) );
        String content = replacer.replaceProperties( element.getAttribute( "content" ) );
        String skip = replacer.replaceProperties( element.getAttribute( "skip" ) );
        Map params = parseParams( DOMUtils.getChildElementsByTagName( element, "param" ) );

        String paramLogsUrl = (String) params.get( "logs.url" );
        if ( paramLogsUrl != null )
        {
            if ( url == null )
            {
                url = paramLogsUrl;
                addWarning( "in output element, param logs.url is now deprecated: please use url attribute (<output url=\"...\">)" );
            }
            else
            {
                addWarning( "in output element, attribute url and deprecated param logs.url have both been set: ignoring param" );
            }
        }

        LogDistiller.Report[] reports = parseReports( DOMUtils.getChildElementsByTagName( element, "report" ) );

        return new LogDistiller.Output( directory, url, content, skip, params, reports );
    }

    protected LogDistiller.Report[] parseReports( Iterator iter )
    {
        List reports = new ArrayList();
        while ( iter.hasNext() )
        {
            Element report = (Element) iter.next();
            reports.add( parseReport( report ) );
        }
        return (LogDistiller.Report[]) reports.toArray( new LogDistiller.Report[reports.size()] );
    }

    protected LogDistiller.Report parseReport( Element element )
    {
        String publisher = replacer.replaceProperties( element.getAttribute( "publisher" ) );
        String format = replacer.replaceProperties( element.getAttribute( "format" ) );
        Map params = parseParams( DOMUtils.getChildElementsByTagName( element, "param" ) );
        return new LogDistiller.Report( publisher, format, params );
    }

    protected LogDistiller.Report[] buildCompatibilityReports( Map params )
    {
        LogDistiller.Report fileReport = new LogDistiller.Report( "file", "txt", new HashMap() );
        String to = (String) params.get( "mail.to" );
        if ( to == null )
        {
            return new LogDistiller.Report[] { fileReport };
        }
        Map newParams = new HashMap();
        newParams.put( "to", replacer.replaceProperties( to ) );
        String cc = (String) params.get( "mail.cc" );
        if ( cc != null )
        {
            newParams.put( "cc", replacer.replaceProperties( cc ) );
        }
        LogDistiller.Report mailReport = new LogDistiller.Report( "mail", "txt", newParams );
        return new LogDistiller.Report[] { fileReport, mailReport };
    }

    protected Map parseParams( Iterator iter )
    {
        Map params = new HashMap();
        while ( iter.hasNext() )
        {
            Element param = (Element) iter.next();
            String name = param.getAttribute( "name" );
            String value = replacer.replaceProperties( DOMUtils.getPCDATA( param ) );
            if ( params.put( name, value ) != null )
            {
                addWarning( "multiple values were defined for '" + name
                    + "' parameter: the last one will override other values." );
            }
        }
        return params;
    }

    protected LogDistiller.Category[] parseCategories( Iterator iter )
    {
        List categories = new ArrayList();
        while ( iter.hasNext() )
        {
            Element category = (Element) iter.next();
            String id = category.getAttribute( "id" );
            String description = DOMUtils.getPCDATA( category );
            categories.add( new LogDistiller.Category( id, description ) );
        }
        return (LogDistiller.Category[]) categories.toArray( new LogDistiller.Category[categories.size()] );
    }

    protected LogDistiller.Group[] parseGroups( Iterator iter, LogDistiller.Category[] categories )
    {
        Map mapCategories = new HashMap();
        for ( int i = 0; i < categories.length; i++ )
        {
            LogDistiller.Category category = categories[i];
            mapCategories.put( category.getId(), category );
        }
        List groups = new ArrayList();
        while ( iter.hasNext() )
        {
            Element group = (Element) iter.next();
            groups.add( parseGroup( group, mapCategories ) );
        }
        return (LogDistiller.Group[]) groups.toArray( new LogDistiller.Group[groups.size()] );
    }

    protected LogDistiller.Group parseGroup( Element element, Map categories )
    {
        String id = element.getAttribute( "id" );
        String description = replacer.replaceProperties( DOMUtils.getPCDATAByTagName( element, "description" ) );
        String attr = replacer.replaceProperties( element.getAttribute( "continueProcessing" ) );
        boolean continueProcessing = Boolean.valueOf( ( attr == null ) ? "false" : attr ).booleanValue();
        attr = element.getAttribute( "save" );
        if ( attr == null )
        {
            attr = "true";
        }
        boolean save = !"false".equals( attr );
        Map params = parseParams( DOMUtils.getChildElementsByTagName( element, "param" ) );
        Condition[] conditions = parseConditions( DOMUtils.getChildElementsByTagName( element, "condition" ) );
        LogDistiller.Report[] reports = parseReports( DOMUtils.getChildElementsByTagName( element, "report" ) );
        LogDistiller.Plugin[] plugins = parsePlugins( DOMUtils.getChildElementsByTagName( element, "plugin" ) );
        if ( old )
        {
            reports = buildCompatibilityReports( params );
        }
        LogDistiller.Category category = (LogDistiller.Category) categories.get( element.getAttribute( "category" ) );
        return new LogDistiller.Group( id, description, continueProcessing, save, params, conditions, reports, plugins,
                                       category );
    }

    protected Condition[] parseConditions( Iterator iter )
    {
        List conditions = new ArrayList();
        while ( iter.hasNext() )
        {
            Element condition = (Element) iter.next();
            conditions.add( parseCondition( condition ) );
        }
        return (Condition[]) conditions.toArray( new Condition[conditions.size()] );
    }

    protected Condition parseCondition( Element element )
    {
        String tags = element.getAttribute( "tags" );
        Match[] matches = parseMatchs( DOMUtils.getChildElementsByTagName( element, "match" ) );
        return new Condition( tags, matches );
    }

    protected Match[] parseMatchs( Iterator iter )
    {
        List matchs = new ArrayList();
        while ( iter.hasNext() )
        {
            Element match = (Element) iter.next();
            matchs.add( parseMatch( match ) );
        }
        return (Match[]) matchs.toArray( new Match[matchs.size()] );
    }

    protected Match parseMatch( Element element )
    {
        String attribute = element.getAttribute( "attribute" );
        String type = element.getAttribute( "type" );
        String reference = replacer.replaceProperties( DOMUtils.getPCDATA( element ) );
        return new Match( attribute, type, reference );
    }

    protected LogDistiller.Plugin[] parsePlugins( Iterator iter )
    {
        List plugins = new ArrayList();
        while ( iter.hasNext() )
        {
            Element plugin = (Element) iter.next();
            plugins.add( parsePlugin( plugin ) );
        }
        return (LogDistiller.Plugin[]) plugins.toArray( new LogDistiller.Plugin[plugins.size()] );
    }

    protected LogDistiller.Plugin parsePlugin( Element element )
    {
        String type = element.getAttribute( "type" );
        if ( Plugins.getPlugin( type ) == null )
        {
            throw new PluginConfigException( "plugin type '" + type + "' unknown, valid values: "
                + Plugins.listAllPluginIds() );
        }
        String attr = element.getAttribute( "globalReport" );
        boolean globalReport = Boolean.valueOf( ( attr == null ) ? "true" : attr ).booleanValue();
        attr = element.getAttribute( "groupReport" );
        boolean groupReport = Boolean.valueOf( ( attr == null ) ? "true" : attr ).booleanValue();
        Map params = parseParams( DOMUtils.getChildElementsByTagName( element, "param" ) );
        return new LogDistiller.Plugin( type, globalReport, groupReport, params );
    }
}
